/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdlib.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/physics/2d/rigid/2d_rigid_pt2pt_constraint.h"
#include "shared/physics/2d/rigid/2d_rigid_prismatic_constraint.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// needed for game consistency on different machines
#include "../../utils/misc/framerate.h"

// misc
#include "utils.h"
#include <vector>
using namespace std;

#include "math3d.h"

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"
#include "../../utils/igimp/igimp.h"
#include "../../utils/igimp/igimp_utils.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "IGJ2 - checker's hopper";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 3.0, CameraZ = 25.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
bool simulator_paused = true;
Drag_Force    global_drag(0.02f, 0.02f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);
double Time = 0.0, LastTime = 0.0;

char *HopperFileName = "content/physdb/hopper.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;

    int ControlType;

    std::string LevelFileName;

    float torso_friction;
    float upper_leg_friction;
    float lower_leg_friction;
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

// 3d stuff
// @todo move to hopper
parsed_file *Models3D;
matrix_4x4 *Bones;


// game stuff

Rigid_Body *MakeRBBox( float Width, float Height, float Density, Vec2 x, float O, float i_scale = 1.0 )
{
    float Mass = Density * Width * Height;
    float OneOverMass = 1 / Mass;
    float OneOverCMMomentOfInertia = 1 / ((Mass / 12) * (Width * Width + Height * Height));

    Rigid_Body *rb = new Rigid_Body(make_Body_Box(Vec2(-Width/2,-Height/2),Vec2(Width/2,Height/2)),
                                    x,O,
                                    OneOverMass,i_scale*OneOverCMMomentOfInertia);
    assert(rb);
    rb->k_friction = 15;
    return rb;
}
void TeleportRB( Rigid_Body &rb, Vec2 const &x, float O = 0.0, Vec2 const &v = Vec2(0,0), float w = 0.0 )
{
    rb.x = x;
    rb.O = O;
    rb.v = v;
    rb.w = w;
    rb.update_derived_state();
}

struct hopper {

    hopper( void ) : torso(0), hip(0), upper_leg(0), knee(0), lower_leg(0)
        {
            // init the parms
            torso_dims     = Vec2(0.5,0.5);
            upper_leg_dims = Vec2(0.25f,1);
            lower_leg_dims = Vec2(0.15f,1);
            
            init_torso_pos     = Vec2(0, 0);
            init_upper_leg_pos = Vec2(0,-0.75f);
            init_lower_leg_pos = Vec2(0,-1.25f);

            // make the bodies
            torso = MakeRBBox(torso_dims.x(),torso_dims.y(),10,init_torso_pos,0,0);  assert(torso);
            upper_leg = MakeRBBox(upper_leg_dims.x(),upper_leg_dims.y(),0.5,init_upper_leg_pos,0);  assert(upper_leg);

            lower_leg = MakeRBBox(lower_leg_dims.x(),lower_leg_dims.y(),0.5,init_lower_leg_pos,0);  assert(lower_leg);
            // add a small sphere to the lower leg to get rid of stickiness
            std::auto_ptr<Body> pr = lower_leg->m_Pbody;
            std::auto_ptr<Body_Sphere> ps(new Body_Sphere(2*lower_leg_dims.x(),Vec2(0,init_lower_leg_pos.y()+lower_leg_dims.y()/2)));
            std::auto_ptr<Body_List> pl(new Body_List(2));
            pl->add_body(pr);  pl->add_body(ps);
            lower_leg->set_body(pl);

            simulator.add(*torso);
            simulator.add(*upper_leg);
            simulator.add(*lower_leg);

            // we don't want to collide with ourselves
            torso->m_Pbody->m_collision_plane     = 0x00000001;
            upper_leg->m_Pbody->m_collision_plane = 0x00000002;
            lower_leg->m_Pbody->m_collision_plane = 0x00000004;

            // make the joints

            // hip
            float hipy = init_torso_pos.y() - init_upper_leg_pos.y();
            hip = new Pt2Pt_Constraint(Handle(*torso,Vec2(0,0)),
                                       Handle(*upper_leg,Vec2(0,hipy)));
            assert(hip);
            simulator.add(*hip);
            hip_spring = new Angle_Spring_Force(Handle(world_body,Vec2(0,0)),
                                                Handle(*upper_leg,Vec2(0,0)),
                                                1000,60,0.0);
            assert(hip_spring);
            simulator.add(*hip_spring);
            

            // knee
            knee = new Prismatic_Constraint(Handle(*upper_leg,Vec2(0,0)),
                                            Handle(*lower_leg,Vec2(0,0)),
                                            Vec2(0,1),Vec2(0,1));
            assert(knee);
            simulator.add(*knee);
            knee_spring = new Spring_Force(Handle(*upper_leg,Vec2(0,5)),
                                           Handle(*lower_leg,Vec2(0,0)),
                                           1000,20,5.0);
            assert(knee_spring);
            simulator.add(*knee_spring);
        }

    void Reset( void )
        {
            assert(torso && upper_leg && lower_leg);
            TeleportRB(*torso,init_torso_pos);
            TeleportRB(*upper_leg,init_upper_leg_pos);
            TeleportRB(*lower_leg,init_lower_leg_pos);
        }

    void Step( void )
        {
            
        }

    void Draw( void )
        {
            if(torso && upper_leg && lower_leg) {
                glColor3f(1,0,0);
                glBody1v(torso->m_Pbody.get(), true);
                glColor3f(0,1,0);
                glBody1v(upper_leg->m_Pbody.get(), true);
                glColor3f(0,0,1);
                glBody1v(lower_leg->m_Pbody.get(), true);
            }
        }
    Vec2 torso_dims, upper_leg_dims, lower_leg_dims;
    Vec2 init_torso_pos, init_upper_leg_pos, init_lower_leg_pos;
    Rigid_Body *torso;
    Pt2Pt_Constraint *hip;
    Angle_Spring_Force *hip_spring;
    Rigid_Body *upper_leg;
    Prismatic_Constraint *knee;
    Spring_Force *knee_spring;
    Rigid_Body *lower_leg;
    
} Hopper;

struct level {

    int num_rb, num_force;
    Rigid_Body **rb_list;
    Binary_Force **force_list;
	float (*colors)[3];
    
    level( void ) : num_rb(0), num_force(0), rb_list(0), force_list(0), colors(0)
        {
        }

    void Load( char const *Filename )
        {            
            // @todo nice, not deleting anything
            // remove the old level
            for(int i=0; i < num_rb; ++i) {
                simulator.remove(*rb_list[i]);
            }
            for(int i=0; i < num_force; ++i) {
                simulator.remove(*force_list[i]);
            }

            // load the new one
            loadPhysicsFile((char*)Filename, 0,0, 1.0/16, 1.0/16,
                            &rb_list, &num_rb, &force_list, &num_force, &world_body);

			float acolors[][3] = {
				0.8,0.2,0.2,
				0.2,0.8,0.2,
				0.2,0.2,0.8,
				0.8,0.8,0.2,
				0.2,0.8,0.8,
				0.8,0.2,0.8,
			};
			int const num_colors = sizeof(acolors)/sizeof(acolors[0]);
			colors = new float[num_rb][3];  assert(colors);
            for(int i=0; i < num_rb; ++i) {
				colors[i][0] = acolors[rand()%num_colors][0];
				colors[i][1] = acolors[rand()%num_colors][1];
				colors[i][2] = acolors[rand()%num_colors][2];
                simulator.add(*rb_list[i]);
            }
            for(int i=0; i < num_force; ++i) {
                simulator.add(*force_list[i]);
            }
        }

    void Draw( void )
        {
            glDisable(GL_DEPTH_TEST);
            for(int i=0; i < num_rb; ++i) {
				glColor3fv(colors[i]);
                glBody1v(rb_list[i]->m_Pbody.get(),true);
                glColor3f(0,0,0);
                glBody1v(rb_list[i]->m_Pbody.get(),false);
            }
            for(int i=0; i < num_force; ++i) {
                glForce1v(force_list[i]);
            }
            glEnable(GL_DEPTH_TEST);
        }
    
} Level;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}


/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
    InitializeFrameRate(120.0f);
    
	PS2Controller.init();

    Globals.Gravity = -20.0f;
    Globals.ControlType = 0;
    Globals.LevelFileName = "hopperquest.physdb";
    Globals.torso_friction = 8.0;
    Globals.upper_leg_friction = 0.0;
    Globals.lower_leg_friction = 10.0;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_CW);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

    char Filename[200];
    sprintf(Filename,"content/physdb/%s",Globals.LevelFileName.c_str());
    Level.Load(Filename);

    // load our 3d crap
    Models3D = ParseIG3DFile("content/3d/alien_hopper.igj3d");
    assert(Models3D && !Models3D->Error && Models3D->FirstSkeleton && !Models3D->FirstSkeleton->Next);
    Models3D->Scale *= 8;
    Bones = new matrix_4x4[Models3D->FirstSkeleton->BoneCount];  assert(Bones);
    
    for(parsed_mesh *mesh = Models3D->FirstMesh;mesh;mesh = mesh->Next) {
        assert(mesh);
        for(int i = 0;i < mesh->GroupCount;++i) {
            parsed_triangle_group &group = mesh->Groups[i];
            char filename[256];
            sprintf(filename,"content/3d/%s",GetFilenameWithoutPath(group.DiffuseTextureFileName));
            group.UserData[0] = LoadTexture(filename);  assert(group.UserData[0]);
        }
        mesh->BoundBones = BindMesh(mesh,Models3D->FirstSkeleton);  assert(mesh->BoundBones);
    }
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    Time = GetTime();
    double dt = Time - LastTime;
    
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

    gravity.m_g.set_y(Globals.Gravity);
    Hopper.torso->k_friction = Globals.torso_friction;
    Hopper.upper_leg->k_friction = Globals.upper_leg_friction;
    Hopper.lower_leg->k_friction = Globals.lower_leg_friction;
    
	PS2Controller.sample();
	Vec2 LeftStick = PS2Controller.left_stick();
	Vec2 RightStick = PS2Controller.right_stick();

    Hopper.knee_spring->m_rest_len = PS2Controller.button_is_down(Joystick::BUTTON_RIGHT_L)?9.0:6.0;
    if(Globals.ControlType == 0) {
        Hopper.hip_spring->m_rest_len = sgn(LeftStick.x())*pow(abs(LeftStick.x()),2.5f)*0.4*M_PI;
    } else {
        if(LeftStick.y() == 0.0 && LeftStick.x() == 0.0) {
            Hopper.hip_spring->m_rest_len = 0.0;
        } else {
            Hopper.hip_spring->m_rest_len = atan2(LeftStick.y(),LeftStick.x()) - M_PI/2;
        }
    }

    if(PS2Controller.button_went_down(Joystick::BUTTON_RIGHT_D)) {
        Hopper.Reset();
    }
    if(PS2Controller.button_went_down(Joystick::BUTTON_START)) {
        simulator_paused = !simulator_paused;
    }
    
	/**************************** step the simulator ****************************/   

    if (spring)                             //update
        spring->m_a = Handle(world_body, world_mouse);

    if(!simulator_paused) {
        int const NumberOfSimSteps = 2;
        for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
            // step as many times as you like
            simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
            // do any game stuff you want here (like per-step collision response, etc.)
        }
        // now wait for the rest of the time so we're consistent on fast machines
        WaitFrameRate(GetFrameRate());
    } else {
        simulator.update_bodies_derived_state();
    }
    

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    CameraX = Hopper.torso->x.x();
    CameraY = Hopper.torso->x.y();
//    CameraZ = 40; // 15 + 50*0.5*(tanh(5*Hopper.torso->v.len()/50-3)+1);
        
    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();

    // draw mouse
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();

    glForce1v(spring);

    Hopper.Draw();
    Level.Draw();

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER,0.5);

#if 0
#if 1
    glEnable(GL_TEXTURE_2D);
    glColor3f(1,1,1);
    glPushMatrix();
    glTranslatef(Hopper.torso->x.x(),Hopper.torso->x.y(),0);
    glRotatef(90,0,1,0);
    glScalef(1/Models3D->Scale,1/Models3D->Scale,1/Models3D->Scale);
    for(int i = 0;i < Models3D->FirstSkeleton->BoneCount;++i) {
        CreateOpenGLTransform(matrix_3x3(matrix_3x3::Identity),vector_3(0,0,0),(float*)Bones[i]);
    }
    for(parsed_mesh *mesh = Models3D->FirstMesh;mesh;mesh = mesh->Next) {
        for(int i = 0;i < mesh->GroupCount;++i) {
            parsed_triangle_group &group = mesh->Groups[i];
            glBindTexture(GL_TEXTURE_2D,group.UserData[0]);
            DrawMeshTriGroup(mesh,Models3D->FirstSkeleton,mesh->BoundBones,(float *)Bones,group);
        }
    }
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);
#else
    glPushMatrix();
    glScalef(1/Models3D->Scale,1/Models3D->Scale,1/Models3D->Scale);
    {for(parsed_mesh *Mesh = Models3D->FirstMesh;
         Mesh;
         Mesh = Mesh->Next)
    {
        glColor3f(0, 0, 0);
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glBegin(GL_TRIANGLES);
        {for(int Index = 0;
             Index < 3*Mesh->TriangleCount;
             ++Index)
        {
            glVertex3fv(Mesh->Vertices[Mesh->Indices[Index]]);
        }}
        glEnd();
    }}
    glPopMatrix();
#endif
#endif  
    glDisable(GL_ALPHA_TEST);
    glDisable(GL_BLEND);
    
    glDisable(GL_BLEND);

    glColor3f(0, 0, 0);
    char buf[100];
    sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
    DrawString(50,50,buf);

    DrawTimingBars();
    SwapBuffers(WindowDC);
    LastTime = Time;
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_PAUSE: {
                    simulator_paused = !simulator_paused;
                    break;
                }
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'H': {
                    Hopper.Reset();
                } break;
                case 'L': {
		            char Filename[200];
		            sprintf(Filename,"content/physdb/%s",Globals.LevelFileName.c_str());
                    launchEditor(Filename);
                    Level.Load(Filename);
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
			if(MovingCamera)
			{
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
				CameraX -= d_world.x();
				CameraY -= d_world.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
