/*
  Indie Game Jam 0 (IGJ0) - http://www.indiegamejam.com
  Copyright 2002 Chris Hecker, Sean Barrett, Casey Muratori, Jonathan Blow, Doug Church

  This file is part of IGJ0.

  IGJ0 is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  IGJ0 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with IGJ0 (in the file COPYING); if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* ========================================================================
   Source File: d6math3d.cpp
   Description: Yet another cheesy 3-vector, 3x3-matrix library
   ======================================================================== */

/* ========================================================================
   Explicit Dependencies
   ======================================================================== */
#include <assert.h>
#include "math3d.h"


//*** We're intentionally converting doubles to floats all over...
#pragma warning ( disable : 4136 )
#pragma warning ( disable : 4244 )

/*----------------------------------------------------------------------------

matrix multiplication.

*/

matrix_3x3 operator*( matrix_3x3 const &Multiplicand,
                      matrix_3x3 const &Multiplier )
{
	matrix_3x3 ReturnMatrix;

	for(int i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			float32 Value = 0;
		  
			for(int k = 0;k < 3;k++)
			{
				Value += Multiplicand.GetElement(i,k) *
               Multiplier.GetElement(k,j);
			}

			ReturnMatrix.SetElement(i,j,Value);
		}
	}

	return ReturnMatrix;
}

vector_3 operator*( matrix_3x3 const &Multiplicand,
                    vector_3 const &Multiplier )
{
	vector_3 ReturnPoint;

	for(int i = 0;i < 3;i++)
	{
		float32 Value = 0;
	  
		for(int k = 0;k < 3;k++)
		{
			Value += Multiplicand.GetElement(i,k) *
            Multiplier.GetElement(k);
		}

		ReturnPoint.SetElement(i,Value);
	}

	return ReturnPoint;
}


/*----------------------------------------------------------------------------

constructors.

*/

matrix_3x3::matrix_3x3( void )
{
   InitIdentity();
}

void matrix_3x3::InitIdentity( void )
{
	for(int Counter = 0;Counter < 9;Counter++)
	{
		aElements[0][Counter] = 0;
	}

	aElements[0][0] = aElements[1][1] = aElements[2][2] = 1;
}

void matrix_3x3::InitWith( vector_3 const &Vector, float32 Cos, float32 Sin )
{
	float32 Vx = Vector.GetX(), Vx2 = Vx * Vx;
	float32 Vy = Vector.GetY(), Vy2 = Vy * Vy;
	float32 Vz = Vector.GetZ(), Vz2 = Vz * Vz;

	aElements[0][0] = Vx2 + Cos * (1 - Vx2);
	aElements[0][1] = Vx * Vy * (1 - Cos) - Vz * Sin;
	aElements[0][2] = Vz * Vx * (1 - Cos) + Vy * Sin;
	
	aElements[1][0] = Vx * Vy * (1 - Cos) + Vz * Sin;
	aElements[1][1] = Vy2 + Cos * (1 - Vy2);
	aElements[1][2] = Vy * Vz * (1 - Cos) - Vx * Sin;
	
	aElements[2][0] = Vz * Vx * (1 - Cos) - Vy * Sin;
	aElements[2][1] = Vy * Vz * (1 - Cos) + Vx * Sin;
	aElements[2][2] = Vz2 + Cos * (1 - Vz2);
}

void matrix_3x3::InitWith( vector_3 const &Vector, float32 Radians )
{
	float32 Sin = sin(Radians), Cos = cos(Radians);
   InitWith(Vector,Cos,Sin);
}

void matrix_3x3::InitWith( vector_3 const &XAxis, vector_3 const &YAxis,
                           vector_3 const &ZAxis )
{
	// put in rotation

	int Counter;
	for(Counter = 0;Counter < 3;Counter++)
	{
		SetElement(0,Counter,XAxis.GetElement(Counter));
		SetElement(1,Counter,YAxis.GetElement(Counter));
		SetElement(2,Counter,ZAxis.GetElement(Counter));
	}
}	
	
/*----------------------------------------------------------------------------

Rotations.

*/

matrix_3x3 &matrix_3x3::ConcatenateXRotation( float32 Radians )
{
	float32 Temp10, Temp11, Temp12;
	float32 Temp20, Temp21, Temp22;

	float32 Sin = sin(Radians), Cos = cos(Radians);

	Temp10 = aElements[1][0] * Cos + aElements[2][0] * -Sin;
	Temp11 = aElements[1][1] * Cos + aElements[2][1] * -Sin;
	Temp12 = aElements[1][2] * Cos + aElements[2][2] * -Sin;

	Temp20 = aElements[1][0] * Sin + aElements[2][0] * Cos;
	Temp21 = aElements[1][1] * Sin + aElements[2][1] * Cos;
	Temp22 = aElements[1][2] * Sin + aElements[2][2] * Cos;

	aElements[1][0] = Temp10;
	aElements[1][1] = Temp11;
	aElements[1][2] = Temp12;
	aElements[2][0] = Temp20;
	aElements[2][1] = Temp21;
	aElements[2][2] = Temp22;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateYRotation( float32 Radians )
{
	float32 Temp00, Temp01, Temp02;
	float32 Temp20, Temp21, Temp22;

	float32 Sin = sin(Radians), Cos = cos(Radians);

	Temp00 = aElements[0][0] * Cos + aElements[2][0] * Sin;
	Temp01 = aElements[0][1] * Cos + aElements[2][1] * Sin;
	Temp02 = aElements[0][2] * Cos + aElements[2][2] * Sin;

	Temp20 = aElements[0][0] * -Sin + aElements[2][0] * Cos;
	Temp21 = aElements[0][1] * -Sin + aElements[2][1] * Cos;
	Temp22 = aElements[0][2] * -Sin + aElements[2][2] * Cos;

	aElements[0][0] = Temp00;
	aElements[0][1] = Temp01;
	aElements[0][2] = Temp02;
	aElements[2][0] = Temp20;
	aElements[2][1] = Temp21;
	aElements[2][2] = Temp22;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateZRotation( float32 Radians )
{
	float32 Temp00, Temp01, Temp02;
	float32 Temp10, Temp11, Temp12;

	float32 Sin = sin(Radians), Cos = cos(Radians);

	Temp00 = aElements[0][0] * Cos + aElements[1][0] * -Sin;
	Temp01 = aElements[0][1] * Cos + aElements[1][1] * -Sin;
	Temp02 = aElements[0][2] * Cos + aElements[1][2] * -Sin;

	Temp10 = aElements[0][0] * Sin + aElements[1][0] * Cos;
	Temp11 = aElements[0][1] * Sin + aElements[1][1] * Cos;
	Temp12 = aElements[0][2] * Sin + aElements[1][2] * Cos;

	aElements[0][0] = Temp00;
	aElements[0][1] = Temp01;
	aElements[0][2] = Temp02;
	aElements[1][0] = Temp10;
	aElements[1][1] = Temp11;
	aElements[1][2] = Temp12;

	return *this;
}

/*----------------------------------------------------------------------------

Scales.

*/

matrix_3x3 &matrix_3x3::ConcatenateXScale( float32 Distance )
{
	aElements[0][0] = aElements[0][0] * Distance;
	aElements[0][1] = aElements[0][1] * Distance;
	aElements[0][2] = aElements[0][2] * Distance;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateYScale( float32 Distance )
{
	aElements[1][0] = aElements[1][0] * Distance;
	aElements[1][1] = aElements[1][1] * Distance;
	aElements[1][2] = aElements[1][2] * Distance;

	return *this;
}

matrix_3x3 &matrix_3x3::ConcatenateZScale( float32 Distance )
{
	aElements[2][0] = aElements[2][0] * Distance;
	aElements[2][1] = aElements[2][1] * Distance;
	aElements[2][2] = aElements[2][2] * Distance;

	return *this;
}

/*----------------------------------------------------------------------------

vector length and normalize.

*/

float32 vector_3::Length( void ) const
{
	return sqrt(GetX()*GetX() + GetY()*GetY() + GetZ()*GetZ());
}

vector_3 &vector_3::Normalize( float32 *OutLength )
{
	float32 MyLength = Length();
   if(OutLength) {
      *OutLength = MyLength;
   }
	if(MyLength != 0)
	{
		SetX(GetX() / MyLength);
		SetY(GetY() / MyLength);
		SetZ(GetZ() / MyLength);
	}

	return *this;
}
  
float32 AngleBetween( vector_3 const &Operand1, vector_3 const &Operand2 )
{
	vector_3 Unit1 = Operand1, Unit2 = Operand2;

	Unit1.Normalize();
	Unit2.Normalize();
	
	return acos(DotProduct(Unit1,Unit2));
}

float32 AngleBetweenUnit( vector_3 const &Unit1, vector_3 const &Unit2 )
{
	return acos(DotProduct(Unit1,Unit2));
}

void GetDextralCoordinateSystem( vector_3 const &InputX,
                                 vector_3 &OutputY, vector_3 &OutputZ )
{
	// use algorithm from Shabana, p. 39

	vector_3 Perp(1,1,1);

	float32 Max = 0;
	int Index = 0;

	for(int i = 0;i < 3;i++)
	{
		float32 Value = fabs(InputX.GetElement(i));
		if(Value > Max)
		{
			Max = Value;
			Index = i;
		}
	}

	Perp.SetElement(Index,0);

	OutputY = CrossProduct(Perp,InputX).Normalize();
	OutputZ = CrossProduct(InputX,OutputY).Normalize();

   float32 const Epsilon = float32(1e-6);
	assert(fabs(InputX.Length() - 1) < Epsilon);
	assert(fabs(OutputY.Length() - 1) < Epsilon);
	assert(fabs(OutputZ.Length() - 1) < Epsilon);
	assert(fabs(DotProduct(InputX,OutputY)) < Epsilon);
	assert(fabs(DotProduct(InputX,OutputZ)) < Epsilon);
	assert(fabs(DotProduct(OutputZ,OutputY)) < Epsilon);
}


void OrthonormalizeOrientation( matrix_3x3 &Orientation )
{
	vector_3 X(Orientation(0,0),Orientation(1,0),Orientation(2,0));
	vector_3 Y(Orientation(0,1),Orientation(1,1),Orientation(2,1));
	vector_3 Z;

	X.Normalize();
	Z = CrossProduct(X,Y).Normalize();
	Y = CrossProduct(Z,X).Normalize();

	Orientation(0,0) = X(0); Orientation(0,1) = Y(0); Orientation(0,2) = Z(0);
	Orientation(1,0) = X(1); Orientation(1,1) = Y(1); Orientation(1,2) = Z(1);
	Orientation(2,0) = X(2); Orientation(2,1) = Y(2); Orientation(2,2) = Z(2);
}

float32 RayIntersectsPlane(vector_3 const &ray_origin,
                        vector_3 const &ray_direction,
                        vector_3 const &plane_normal,
                        float32 const plane_d,
                        float32 &vd)
{
	float32 t;

	// Determine the angle of incidence between the ray and the plane
	vd = DotProduct(plane_normal, ray_direction);
	
	if(fabs(vd) > float(1e-6))
	{
		// The rey interesects the plane
		// Compute the intersection distance, t
		float32 vn = DotProduct(plane_normal, ray_origin) + plane_d;
		t = -vn / vd;
	}
	else
	{
		// The ray is parallel, intersects only at infinity
		t = FLOAT32_MAX;
	}

	return(t);
}

bool RayIntersectsBoundingBox(vector_3 const &ray_origin,
                              vector_3 const &ray_direction,
                              vector_3 const &box_origin,
                              matrix_3x3 const &box_orientation,
                              float32 const width,
                              float32 const height,
                              float32 const depth,
                              float32 &t, vector_3 &normal)
{
	float32 const dimension_array[3] = {width/2, height/2, depth/2};

	// Begin with the whole line
	float32 t_near = -FLOAT32_MAX;
	float32 t_far = FLOAT32_MAX;
	
	for(unsigned int counter = 0; counter < 6; counter++)
	{
		unsigned int dimension_index = counter / 2;
		float32 sign = (counter % 2) ? 1 : -1;
		float32 dimension = dimension_array[dimension_index];

		// Determine the plane made by this part of the bounding box
		vector_3 plane_normal(box_orientation.GetElement(0, dimension_index),
                            box_orientation.GetElement(1, dimension_index),
                            box_orientation.GetElement(2, dimension_index));
		plane_normal = plane_normal * sign;
		float32 plane_d = dimension - DotProduct(box_origin,
                                            plane_normal);

		// Intersect the line with the plane
		float32 vd;
		float32 t_test = RayIntersectsPlane(ray_origin, ray_direction,
                                       plane_normal, plane_d, vd);

		// Clip the line to the plane
		if(vd < 0)
		{
			// Plane is a back-face, affects t_far
			if(t_test < t_far)
			{
				t_far = t_test;
			}
		}
		else
		{
			// Plane is a front-face, affects t_near
			if(t_test > t_near)
			{
				normal = -plane_normal;
				t_near = t_test;
			}
		}
	}

	// If we have any line left, we've intersected
	t = t_near;
	return((t_far - t_near) > float32(1e-6));
}

/*----------------------------------------------------------------------------

Inverse

@todo both of these inverses suck ass!!!

*/

matrix_3x3 Inverse( matrix_3x3 const &A )
{
	matrix_3x3 Inv;

	float32 Adet = A(0,0) * (A(1,1)*A(2,2) - A(1,2)*A(2,1)) -
      A(0,1) * (A(1,0)*A(2,2) - A(1,2)*A(2,0)) +
      A(0,2) * (A(1,0)*A(2,1) - A(1,1)*A(2,0));

	assert(Adet != 0);
	
	Inv(0,0) = (A(1,1)*A(2,2) - A(1,2)*A(2,1))/Adet;
	Inv(1,0) = -(A(1,0)*A(2,2) - A(1,2)*A(2,0))/Adet;
	Inv(2,0) = (A(1,0)*A(2,1) - A(1,1)*A(2,0))/Adet;

	Inv(0,1) = -(A(0,1)*A(2,2) - A(0,2)*A(2,1))/Adet;
	Inv(1,1) = (A(0,0)*A(2,2) - A(0,2)*A(2,0))/Adet;
	Inv(2,1) = -(A(0,0)*A(2,1) - A(0,1)*A(2,0))/Adet;

	Inv(0,2) = (A(0,1)*A(1,2) - A(0,2)*A(1,1))/Adet;
	Inv(1,2) = -(A(0,0)*A(1,2) - A(0,2)*A(1,0))/Adet;
	Inv(2,2) = (A(0,0)*A(1,1) - A(0,1)*A(1,0))/Adet;

	return Inv;
}

bool Inverse( matrix_3x3 const &A, matrix_3x3 &Inv )
{
	bool Inverted = false;

	float32 Adet = A(0,0) * (A(1,1)*A(2,2) - A(1,2)*A(2,1)) -
      A(0,1) * (A(1,0)*A(2,2) - A(1,2)*A(2,0)) +
      A(0,2) * (A(1,0)*A(2,1) - A(1,1)*A(2,0));

	// @todo need to do something with the condition number or norms here
	if(Adet != 0)
	{
		Inv(0,0) = (A(1,1)*A(2,2) - A(1,2)*A(2,1))/Adet;
		Inv(1,0) = -(A(1,0)*A(2,2) - A(1,2)*A(2,0))/Adet;
		Inv(2,0) = (A(1,0)*A(2,1) - A(1,1)*A(2,0))/Adet;
	
		Inv(0,1) = -(A(0,1)*A(2,2) - A(0,2)*A(2,1))/Adet;
		Inv(1,1) = (A(0,0)*A(2,2) - A(0,2)*A(2,0))/Adet;
		Inv(2,1) = -(A(0,0)*A(2,1) - A(0,1)*A(2,0))/Adet;
	
		Inv(0,2) = (A(0,1)*A(1,2) - A(0,2)*A(1,1))/Adet;
		Inv(1,2) = -(A(0,0)*A(1,2) - A(0,2)*A(1,0))/Adet;
		Inv(2,2) = (A(0,0)*A(1,1) - A(0,1)*A(1,0))/Adet;

		Inverted = true;
	}

	return Inverted;
}

bool Solve3x3InPlace( matrix_3x3 &A, vector_3 &b )
{
   // gaussian elimination with row pivoting
   // @optimize could MGS QR be faster because no pivoting?
   // @optimize or just totally unroll this, or is this compiler doing that?
   int i, j, k;
   // forward elminate
   for(i = 0;i < 3;++i) {
      // find the largest pivot in this column
      float32 maxabs = FLOAT32_MIN;
      int pivot_idx = 3;
      for(j = i;j < 3;++j) {
         float32 el = fabs(A.aElements[j][i]);
         if(el > maxabs) {
            maxabs = el;
            pivot_idx = j;
         }
      }
         
      // swap the rows
      if(pivot_idx >= 3) {
         return false;
      }
      if(pivot_idx != i) { 
         for(j = i;j < 3;++j) {
            float32 t = A.aElements[pivot_idx][j];
            A.aElements[pivot_idx][j] = A.aElements[i][j];
            A.aElements[i][j] = t;
         }
         float32 t = b(pivot_idx);
         b(pivot_idx) = b(i);
         b(i) = t;
      }
      
      // pivot the row
      float32 pivot = A.aElements[i][i];
      if(fabs(pivot) < 1e-6) {
         return false;
      }
      float32 inv_pivot = 1.0f/pivot;
      for(j = i+1;j < 3;++j) {
         A.aElements[i][j] *= inv_pivot;
      }
      b(i) *= inv_pivot;

      // eliminate the trailing submatrix
      for(j = i+1;j < 3;++j) {
         float32 c = A.aElements[j][i];
         for(k = i+1;k < 3;++k) {
            A.aElements[j][k] -= c * A.aElements[i][k];
         }
         b(j) -= c * b(i);
      }
   }
   // backwards substitute
   // @optimize x(2) is already in b(2)
   for(i = 2;i >= 0;--i) {
      for(j = 2;j > i;--j) {
         b(i) -= A.aElements[i][j] * b(j);
      }
   }
   return true;
}

#include <stdio.h>

void Print( vector_3 const &a )
{
   printf("{ %10.5f, %10.5f, %10.5f }\n",a.x,a.y,a.z);
}
void Print( matrix_3x3 const &a )
{
   printf("{{ %10.5f, %10.5f, %10.5f },\n",a.aElements[0][0],a.aElements[0][1],a.aElements[0][2]);
   printf(" { %10.5f, %10.5f, %10.5f },\n",a.aElements[1][0],a.aElements[1][1],a.aElements[1][2]);
   printf(" { %10.5f, %10.5f, %10.5f }}\n",a.aElements[2][0],a.aElements[2][1],a.aElements[2][2]);
}

//*** Turn on float conversion warning
#pragma warning ( default : 4136 )
#pragma warning ( disable : 4514 ) // unused inline function


