/*
  Indie Game Jam 0 (IGJ0) - http://www.indiegamejam.com
  Copyright 2002 Chris Hecker, Sean Barrett, Casey Muratori, Jonathan Blow, Doug Church

  This file is part of IGJ0.

  IGJ0 is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  IGJ0 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with IGJ0 (in the file COPYING); if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#if !defined(MATH3D_H)
/* ========================================================================
   Header File: d6math3d.h
   Description: Yet another cheesy 3-vector, 3x3-matrix library,
   transmogrified from dumb3d.h
   ======================================================================== */

/* ========================================================================
   Explicit Dependencies
   ======================================================================== */
#include <math.h>		// for sqrt
#include <float.h>      // for min/max
#include <stdlib.h>     // rand & RAND_MAX
#include <assert.h>

typedef float            float32;
typedef double           float64;
typedef char      signed    int8;
typedef char    unsigned   uint8;
typedef short              int16;
typedef short   unsigned  uint16;
typedef int                int32;
typedef int     unsigned  uint32;
typedef __int64            int64;
typedef __int64 unsigned  uint64;

// The default real type
#define FLOAT32_IS_FLOAT  1
#define FLOAT32_IS_DOUBLE 0
typedef float float32;
float32 const FLOAT32_MAX = FLT_MAX;
float32 const FLOAT32_MIN = FLT_MIN;
float32 const Float32Pi = float32(3.14159265358979323846);
inline float32 sqr( float32 x ) { return x*x; }
inline float32 sgn( float32 x ) { return x >= 0 ? 1.0f : -1.0f; }

/* Remove MacIntosh lameness */
#if defined(Length)
#undef Length
#endif

#pragma warning (disable:4244)		// int to float conversion

/*----------------------------------------------------------------------------

vector_3.  This class represents a 3D vector.

*/

class vector_3
{
public:

	inline vector_3( void );
	inline vector_3( float32 const *pElements );
	inline vector_3( float32 X, float32 Y, float32 Z );
    inline vector_3( vector_3 const & );
    inline vector_3 &operator=( vector_3 const & );

	inline float32 &operator()( int unsigned Row );
	inline float32 const &operator()( int unsigned Row ) const;

	inline float32 GetElement( int Row ) const;

	inline float32 GetX( void ) const;
	inline float32 GetY( void ) const;
	inline float32 GetZ( void ) const;

	inline void SetElement( int Row, float32 Value );

	inline void SetX( float32 Value );
	inline void SetY( float32 Value );
	inline void SetZ( float32 Value );

    inline float32 *p(void) {return aElements;}
	inline float32 *GetElementsPointer(void) {return(aElements);}
	inline float32 *GetElements(void) {return(aElements);}

	inline vector_3 &operator+=(vector_3 const &addend);
	inline vector_3 &operator-=(vector_3 const &lessor);
	inline vector_3 &operator*=(float32 const multiplier);
	inline vector_3 &operator/=(float32 const divisor);

	inline bool IsZero( void ) const;
	vector_3 &Normalize( float32 *Length = 0 );
	float32 Length( void ) const;

   union {
      struct {
         float32 x, y, z;
      };
      float32 aElements[3];
   };

	enum do_not_initialize { DoNotInitialize };
	inline vector_3( do_not_initialize ) {};
};



/*----------------------------------------------------------------------------

matrix_3x3.  This class represents row major 3x3 homogeneous matrices.

*/

class matrix_3x3
{
public:

	matrix_3x3( void );
	matrix_3x3( float32 (*pElements)[3] );
	matrix_3x3( float32 const pElements[9] );

	// @todo this isn't the safest ctor to have
	enum mode { Identity, SkewSymmetric, Zero, Rows, Columns, Diagonal, OuterProduct,
               XRotation, YRotation, ZRotation };
	inline matrix_3x3( mode MustBeZeroOrIdentity );
	inline matrix_3x3( vector_3 const &CrossVector, mode MustBeSkewSymmetric );
	inline matrix_3x3( vector_3 const &a, vector_3 const &b,
                      vector_3 const &c, mode MustBeRowsOrColumns );
	inline matrix_3x3( float32 const d, mode MustBeDiagonalOrXYZRotation );
	inline matrix_3x3( vector_3 const &a, vector_3 const &b,
                      mode MustBeOuterProduct );
	
	inline matrix_3x3( vector_3 const &Axis, float32 AngleInRadians );
	inline matrix_3x3( vector_3 const &Axis, float32 CosAngleInRadians, float32 SinAngleInRadians );
	inline matrix_3x3( vector_3 const &XAxis, vector_3 const &YAxis,
                      vector_3 const &ZAxis);
	inline matrix_3x3( matrix_3x3 const &Rotation );

	matrix_3x3 &ConcatenateXRotation( float32 Radians );
	matrix_3x3 &ConcatenateYRotation( float32 Radians );
	matrix_3x3 &ConcatenateZRotation( float32 Radians );

	matrix_3x3 &ConcatenateXScale( float32 Distance );
	matrix_3x3 &ConcatenateYScale( float32 Distance );
	matrix_3x3 &ConcatenateZScale( float32 Distance );
	inline matrix_3x3 &ConcatenateScale( vector_3 &Scale );

	inline matrix_3x3 &operator+=( matrix_3x3 const &A );
	inline matrix_3x3 &operator*=( matrix_3x3 const &A );
	inline matrix_3x3 &operator-=( matrix_3x3 const &A );

	inline float32 &operator()( int unsigned Row, int unsigned Column );
	inline float32 const &operator()( int unsigned Row, int unsigned Column ) const;

	inline float32 GetElement( int Row, int Column ) const;
	inline matrix_3x3 &SetElement( int Row, int Column, float32 Value );

	// @todo we need a row vector class
	inline vector_3 GetRow( int unsigned Row ) const;
	inline vector_3 GetColumn( int unsigned Column ) const;

   inline float32 *p(void) {return aElements[0];}
	inline float32 const *GetElementsPointer( void ) const { return aElements[0]; }

	inline bool IsIdentity( void ) const;
	inline bool IsZero( void ) const;

	enum do_not_initialize { DoNotInitialize };

	inline matrix_3x3( do_not_initialize );

   void InitIdentity( void );
	void InitWith( vector_3 const &Axis, float32 AngleInRadians );
	void InitWith( vector_3 const &Axis, float32 CosAngleInRadians, float32 SinAngleInRadians );
	void InitWith( vector_3 const &XAxis, vector_3 const &YAxis,
                  vector_3 const &ZAxis );

   union {
      float32 aElements9[9];
      float32 aElements[3][3];
   };
};

/*----------------------------------------------------------------------------

globally useful noninline functions.

*/

void Print( vector_3 const &a );
void Print( matrix_3x3 const &a );

bool RayIntersectsBoundingBox(vector_3 const &ray_origin,
                              vector_3 const &ray_direction,
                              vector_3 const &box_origin,
                              matrix_3x3 const &box_orientation,
                              float32 const width,
                              float32 const height,
                              float32 const depth,
                              float32 &t, vector_3 &normal);

float32 RayIntersectsPlane(vector_3 const &ray_origin,
                        vector_3 const &ray_direction,
                        vector_3 const &plane_normal,
                        float32 const plane_d,
                        float32 &vd);

matrix_3x3 Inverse( matrix_3x3 const &A );
bool Inverse( matrix_3x3 const &A, matrix_3x3 &AInverse );
bool Solve3x3InPlace( matrix_3x3 &A, vector_3 &b );

void OrthonormalizeOrientation( matrix_3x3 &Orientation );

float32 AngleBetween( vector_3 const &Operand1, vector_3 const &Operand2 );
float32 AngleBetweenUnit( vector_3 const &Unit1, vector_3 const &Unit2 );

void GetDextralCoordinateSystem( vector_3 const &InputX,
                                 vector_3 &OutputY, vector_3 &OutputZ );

matrix_3x3 operator*( matrix_3x3 const &Multiplicand,
                      matrix_3x3 const &Multiplier );

vector_3 operator*( matrix_3x3 const &Multiplicand,
                    vector_3 const &Multiplier );

/*----------------------------------------------------------------------------

inline function definitions.

*/

inline int CorrectModi( int num, int den )
{ // compute the "correct" mod for signed numerator, C default is broken for float32 math
   assert(den > 0);
   if(num >= 0) {
      return num % den;
   } else {
      int m = (-num) % den;
      if(m != 0) {
         m = den - m;
      }
      return m;
   }
}
   
inline void CorrectModf( float32 a, float32 &Integer, float32 &Fraction )
{ // compute the "correct" mathematical mod, where the fractional part is always >= 0
   double i;
   double f = modf(a,&i);
   if(a >= 0) {
      Integer = i;
      Fraction = f;
   } else {
      // a < 0 needs special treatment
      Integer = i-1;
      Fraction = 1-f;
   }
}
inline void CorrectModfv( vector_3 const &v, vector_3 &Integer, vector_3 &Fraction )
{ 
   for(int i = 0;i < 3;++i) {
      CorrectModf(v(i),Integer(i),Fraction(i));
   }
}

inline vector_3 ceilv( vector_3 const &a )
{
   return vector_3(ceil(a.x),ceil(a.y),ceil(a.z));
}
inline vector_3 floorv( vector_3 const &a )
{
   return vector_3(floor(a.x),floor(a.y),floor(a.z));
}

inline float32 rabs( float32 x ) { return (float32)fabs(x); }

inline float32 RadiansFrom( float32 Degrees )
{
	return (Degrees / float32(360)) * Float32Pi * float32(2);
}

inline float32 DegreesFrom( float32 Radians )
{
	return float32(360) * (Radians / (Float32Pi * float32(2)));
}

inline vector_3 &vector_3::operator+=(vector_3 const &addend)
{
	SetX(GetX() + addend.GetX());
	SetY(GetY() + addend.GetY());
	SetZ(GetZ() + addend.GetZ());

	return(*this);
}

inline vector_3 operator+( vector_3 const &Operand1,
                           vector_3 const &Operand2 )
{
	return vector_3(Operand1.GetX() + Operand2.GetX(),
                   Operand1.GetY() + Operand2.GetY(),
                   Operand1.GetZ() + Operand2.GetZ());
}

inline vector_3 &vector_3::operator-=(vector_3 const &lessor)
{
	SetX(GetX() - lessor.GetX());
	SetY(GetY() - lessor.GetY());
	SetZ(GetZ() - lessor.GetZ());

	return(*this);
}

inline vector_3 operator-( vector_3 const &Operand1,
                           vector_3 const &Operand2 )
{
	return vector_3(Operand1.GetX() - Operand2.GetX(),
                   Operand1.GetY() - Operand2.GetY(),
                   Operand1.GetZ() - Operand2.GetZ());
}

inline vector_3 operator-( vector_3 const &Operand1 )
{
	return vector_3(-Operand1.GetX(),-Operand1.GetY(),-Operand1.GetZ());
}

inline vector_3 &vector_3::operator*=(float32 const multiplier)
{
	SetX(GetX() * multiplier);
	SetY(GetY() * multiplier);
	SetZ(GetZ() * multiplier);

	return(*this);
}

inline vector_3 operator*( vector_3 const &Multiplicand,
                           float32 const &Multiplier )
{
	return vector_3(Multiplicand.GetX() * Multiplier,
                   Multiplicand.GetY() * Multiplier,
                   Multiplicand.GetZ() * Multiplier);
}

inline vector_3 operator*( float32 const &Multiplier,
                           vector_3 const &Multiplicand ) 
{
	return Multiplicand * Multiplier;
}
	  
inline vector_3 &vector_3::operator/=(float32 const divisor)
{
	SetX(GetX() / divisor);
	SetY(GetY() / divisor);
	SetZ(GetZ() / divisor);

	return(*this);
}

inline vector_3 operator/( vector_3 const &Quotient,
                           float32 const &Divisor )
{
	return Quotient * (float32(1.0)/Divisor);
}

inline vector_3 CrossProduct( vector_3 const &Operand1,
                              vector_3 const &Operand2 )
{
	float32 X = Operand1.GetY() * Operand2.GetZ() -
      Operand1.GetZ() * Operand2.GetY();
	float32 Y = Operand1.GetZ() * Operand2.GetX() -
      Operand1.GetX() * Operand2.GetZ();
	float32 Z = Operand1.GetX() * Operand2.GetY() -
      Operand1.GetY() * Operand2.GetX();

	return vector_3(X,Y,Z);
}

inline vector_3 HadamardProduct( vector_3 const &Operand1,
                                 vector_3 const &Operand2 )
{
	return vector_3(Operand1.GetX() * Operand2.GetX(),
                   Operand1.GetY() * Operand2.GetY(),
                   Operand1.GetZ() * Operand2.GetZ());
}

inline float32 DotProduct( vector_3 const &Operand1, vector_3 const &Operand2 )
{
	return Operand1.GetX() * Operand2.GetX() +
      Operand1.GetY() * Operand2.GetY() +
      Operand1.GetZ() * Operand2.GetZ();
}

inline float32 vector_3::GetElement( int Row ) const
{
	return aElements[Row];
}

inline float32 vector_3::GetX( void ) const
{
	return aElements[0];
}

inline float32 vector_3::GetY( void ) const
{
	return aElements[1];
}

inline float32 vector_3::GetZ( void ) const
{
	return aElements[2];
}

inline float32 &vector_3::operator()( int unsigned Row )
{
	assert(Row < 3);
	return aElements[Row];
}

inline float32 const &vector_3::operator()( int unsigned Row ) const
{
	assert(Row < 3);
	return aElements[Row];
}

inline void vector_3::SetElement( int Row, float32 Value )
{
	aElements[Row] = Value;
}

inline void vector_3::SetX( float32 Value )
{
	aElements[0] = Value;
}

inline void vector_3::SetY( float32 Value )
{
	aElements[1] = Value;
}

inline void vector_3::SetZ( float32 Value )
{
	aElements[2] = Value;
}

inline vector_3::vector_3( void )
{
	aElements[0] = aElements[1] = aElements[2] = 0;
}

inline vector_3::vector_3( float32 X, float32 Y, float32 Z )
{
	aElements[0] = X;
	aElements[1] = Y;
	aElements[2] = Z;
}

inline vector_3::vector_3( vector_3 const &Source )
{
	aElements[0] = Source.aElements[0];
	aElements[1] = Source.aElements[1];
	aElements[2] = Source.aElements[2];
}

inline vector_3 &vector_3::operator=( vector_3 const &Source )
{
	aElements[0] = Source.aElements[0];
	aElements[1] = Source.aElements[1];
	aElements[2] = Source.aElements[2];

	return *this;
}

inline vector_3::vector_3( float32 const *pElements )
{
	assert(pElements);
	aElements[0] = pElements[0];
	aElements[1] = pElements[1];
	aElements[2] = pElements[2];
}

inline float32 &matrix_3x3::operator()( int unsigned Row, int unsigned Column )
{
	assert((Row < 3) && (Column < 3));
	return aElements[Row][Column];
}

inline float32 const &matrix_3x3::operator()( int unsigned Row,
                                           int unsigned Column ) const
{
	assert((Row < 3) && (Column < 3));
	return aElements[Row][Column];
}

inline matrix_3x3 &matrix_3x3::operator*=( matrix_3x3 const &A )
{
   *this = *this * A;
   return *this;
}

inline float32 matrix_3x3::GetElement( int Row, int Column ) const
{
	return aElements[Row][Column];
}

inline matrix_3x3 &matrix_3x3::SetElement( int Row, int Column, float32 Value )
{
	aElements[Row][Column] = Value;

	return *this;
}

inline matrix_3x3::matrix_3x3( do_not_initialize )
{
}

inline matrix_3x3::matrix_3x3( vector_3 const &Axis, float32 AngleInRadians )
{
	InitWith(Axis,AngleInRadians);
}
inline matrix_3x3::matrix_3x3( vector_3 const &Axis, float32 CosAngleInRadians, float32 SinAngleInRadians )
{
	InitWith(Axis,CosAngleInRadians,SinAngleInRadians);
}

inline matrix_3x3::matrix_3x3( float32 (*pElements)[3] )
{
	assert(pElements);

	for(int i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			aElements[i][j] = pElements[i][j];
		}
	}
}
inline matrix_3x3::matrix_3x3( float32 const pElements[9] )
{
	assert(pElements);

	for(int i = 0;i < 9;i++)
	{
		aElements[0][i] = pElements[i];
	}
}

inline matrix_3x3::matrix_3x3( vector_3 const &Vector, mode Mode )
{
	assert(Mode == SkewSymmetric);

	matrix_3x3 &M = *this;

	M(0,0) =  0;             M(0,1) = -Vector.GetZ(); M(0,2) =  Vector.GetY();
	M(1,0) =  Vector.GetZ(); M(1,1) =  0;             M(1,2) = -Vector.GetX();
	M(2,0) = -Vector.GetY(); M(2,1) =  Vector.GetX(); M(2,2) = 0;
}

inline matrix_3x3::matrix_3x3( vector_3 const &a, vector_3 const &b, mode Mode )
{
	assert(Mode == OuterProduct);

	matrix_3x3 &M = *this;

	for(int unsigned i = 0;i < 3;i++)
	{
		M(i,0) = a(i)*b(0);
		M(i,1) = a(i)*b(1);
		M(i,2) = a(i)*b(2);
	}
}

inline matrix_3x3::matrix_3x3( mode Mode )
{
	assert((Mode == Zero) || (Mode == Identity));
   switch(Mode) {
      case Zero:
         for(int Counter = 0;Counter < 3;Counter++)
         {
            aElements[Counter][0] = 0;
            aElements[Counter][1] = 0;
            aElements[Counter][2] = 0;
         }
         break;
         
      case Identity:
         InitIdentity();
         break;
   }
}

inline matrix_3x3::matrix_3x3( float32 const d, mode Mode )
{
   switch(Mode)
   {
      case Diagonal:
      {
         for(int Counter = 0;Counter < 3;Counter++)
         {
            aElements[Counter][0] = 0;
            aElements[Counter][1] = 0;
            aElements[Counter][2] = 0;
         }
         aElements[0][0] = aElements[1][1] = aElements[2][2] = d;
         break;
      }
      // @optimize these are dumb
      case XRotation:
      {
         InitIdentity();
         (*this).ConcatenateXRotation(d);
         break;
      }
      case YRotation:
      {
         InitIdentity();
         (*this).ConcatenateYRotation(d);
         break;
      }
      case ZRotation:
      {
         InitIdentity();
         (*this).ConcatenateZRotation(d);
         break;
      }

      default:
      {
         assert(!"mode must be Diagonal or XYZRotation");
         break;
      }
   }
}

inline matrix_3x3::matrix_3x3(vector_3 const &XAxis, vector_3 const &YAxis,
                              vector_3 const &ZAxis )
{
	InitWith(XAxis,YAxis,ZAxis);
}

inline matrix_3x3::matrix_3x3(vector_3 const &a, vector_3 const &b,
                              vector_3 const &c, mode Mode )
{
	assert((Mode == Rows) || (Mode == Columns));

	if(Mode == Columns)
	{
		for(int i = 0;i < 3;i++)
		{
			(*this)(i,0) = a(i);
			(*this)(i,1) = b(i);
			(*this)(i,2) = c(i);
		}
	}
	else
	{
		for(int j = 0;j < 3;j++)
		{
			(*this)(0,j) = a(j);
			(*this)(1,j) = b(j);
			(*this)(2,j) = c(j);
		}
	}
}

inline matrix_3x3::matrix_3x3( matrix_3x3 const &Rotation )
{
	// put in rotation

	for(int Counter = 0;Counter < 3;Counter++)
	{
		SetElement(0,Counter,Rotation.GetElement(0,Counter));
		SetElement(1,Counter,Rotation.GetElement(1,Counter));
		SetElement(2,Counter,Rotation.GetElement(2,Counter));
	}
}

inline matrix_3x3 &matrix_3x3::ConcatenateScale( vector_3 &Scale )
{
	ConcatenateXScale(Scale.GetX());
	ConcatenateYScale(Scale.GetY());
	ConcatenateZScale(Scale.GetZ());

	return *this;
}

inline matrix_3x3 Transpose( matrix_3x3 const &Matrix )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Matrix(Counter,0);
		Return(1,Counter) = Matrix(Counter,1);
		Return(2,Counter) = Matrix(Counter,2);
	}

	return Return;
}	

inline matrix_3x3 operator+( matrix_3x3 const &Operand1,
                             matrix_3x3 const &Operand2 )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Operand1(0,Counter) + Operand2(0,Counter);
		Return(1,Counter) = Operand1(1,Counter) + Operand2(1,Counter);
		Return(2,Counter) = Operand1(2,Counter) + Operand2(2,Counter);
	}

	return Return;
}	
	
inline matrix_3x3 operator-( matrix_3x3 const &Operand1,
                             matrix_3x3 const &Operand2 )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Operand1(0,Counter) - Operand2(0,Counter);
		Return(1,Counter) = Operand1(1,Counter) - Operand2(1,Counter);
		Return(2,Counter) = Operand1(2,Counter) - Operand2(2,Counter);
	}

	return Return;
}	
	
inline matrix_3x3 operator*( float32 const Value, matrix_3x3 const &Matrix )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = Value * Matrix(0,Counter);
		Return(1,Counter) = Value * Matrix(1,Counter);
		Return(2,Counter) = Value * Matrix(2,Counter);
	}

	return Return;
}	

inline matrix_3x3 &matrix_3x3::operator+=( matrix_3x3 const &Operand1 )
{
	for(int Counter = 0;Counter < 3;Counter++)
	{
		// @todo this is probably the wrong order, but I'm lazy
		aElements[0][Counter] += Operand1(0,Counter);
		aElements[1][Counter] += Operand1(1,Counter);
		aElements[2][Counter] += Operand1(2,Counter);
	}

	return *this;
}	

inline matrix_3x3 &matrix_3x3::operator-=( matrix_3x3 const &Operand1 )
{
	for(int Counter = 0;Counter < 3;Counter++)
	{
		// @todo this is probably the wrong order, but I'm lazy
		aElements[0][Counter] -= Operand1(0,Counter);
		aElements[1][Counter] -= Operand1(1,Counter);
		aElements[2][Counter] -= Operand1(2,Counter);
	}

	return *this;
}	

inline matrix_3x3 operator*( matrix_3x3 const &Matrix, float32 const Value )
{
	return Value * Matrix;
}

inline matrix_3x3 operator-( matrix_3x3 const &Matrix )
{
	matrix_3x3 Return;

	for(int Counter = 0;Counter < 3;Counter++)
	{
		Return(0,Counter) = -Matrix(0,Counter);
		Return(1,Counter) = -Matrix(1,Counter);
		Return(2,Counter) = -Matrix(2,Counter);
	}

	return Return;
}	

inline bool operator==( vector_3 const &A, vector_3 const &B)
{
	return((A.GetX() == B.GetX()) &&
          (A.GetY() == B.GetY()) &&
          (A.GetZ() == B.GetZ()));
}

inline vector_3 matrix_3x3::GetRow( int unsigned Row ) const
{
	assert(Row < 3);
	return vector_3((*this)(Row,0),(*this)(Row,1),(*this)(Row,2));
}

inline vector_3 matrix_3x3::GetColumn( int unsigned Column ) const
{
	assert(Column < 3);
	return vector_3((*this)(0,Column),(*this)(1,Column),(*this)(2,Column));
}

inline vector_3 Normalize( vector_3 const &A, float32 *Length = 0 )
{
	vector_3 Return = A;
	return Return.Normalize(Length);
}

inline float32 Length( vector_3 const &A )
{
	return A.Length();
}

inline float32 Length2( vector_3 const &A )
{
	return DotProduct(A,A);
}
inline float32 Distance2( vector_3 const &a, vector_3 const &b )
{
   return sqr(b.x-a.x) + sqr(b.y-a.y) + sqr(b.z-a.z);
}

inline vector_3 &ClipVectorMagnitude( vector_3 &A, float32 const MaximumLength )
{
	if(A.Length() > MaximumLength)
	{
		A.Normalize() *= MaximumLength;
	}
	return A;
}
inline float32 ClipFloat32Magnitude( float32 Value, float32 MaximumMagnitude )
{
	assert(MaximumMagnitude > 0);
	if(rabs(Value) > MaximumMagnitude)
	{
		Value = (Value > 0) ? MaximumMagnitude : -MaximumMagnitude;
	}
	return Value;
}

inline bool vector_3::IsZero( void ) const
{
	return (aElements[0] == 0) &&
      (aElements[1] == 0) &&
      (aElements[2] == 0);
}

inline void CreateOpenGLTransform( matrix_3x3 const &Orientation,
                                   vector_3 const &Position, float *pM)
{
	assert(pM);
	float (*pM44)[4] = (float(*)[4])pM;
	int i;
	
	// remember, OpenGL matrices are transposed!

	for(i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			pM44[j][i] = Orientation(i,j);
		}
	}

	pM44[0][3] = pM44[1][3] = pM44[2][3] = 0.0f; pM44[3][3] = 1.0f;

	for(i = 0;i < 3;i++)
	{
		pM44[3][i] = Position(i);
	}
}

inline void DecomposeOpenGLTransform( float const *pM, matrix_3x3 &Rotation,
                                      vector_3 &Translation )
{
   // @todo what about scale, projection, etc?
	assert(pM);
	float (*pM44)[4] = (float(*)[4])pM;
	int i;
	
	// remember, OpenGL matrices are transposed!

	for(i = 0;i < 3;i++)
	{
		for(int j = 0;j < 3;j++)
		{
			Rotation(i,j) = pM44[j][i];
		}
	}
	for(i = 0;i < 3;i++)
	{
		Translation(i) = pM44[3][i];
	}
}

bool matrix_3x3::IsIdentity( void ) const
{
	float32 const (*p)[3] = aElements;
	
	return  (p[0][0]==1) && (p[0][1]==0) && (p[0][2]==0) &&
      (p[1][0]==0) && (p[1][1]==1) && (p[1][2]==0) &&
      (p[2][0]==0) && (p[2][1]==0) && (p[2][2]==1);
}
	
bool matrix_3x3::IsZero( void ) const
{
	float32 const (*p)[3] = aElements;
   
	return  (p[0][0]==0) && (p[0][1]==0) && (p[0][2]==0) &&
      (p[1][0]==0) && (p[1][1]==0) && (p[1][2]==0) &&
      (p[2][0]==0) && (p[2][1]==0) && (p[2][2]==0);
}

inline float32 ContinuousMaxMagnitude( float32 v, float32 v_max )
{
   return v_max * (float32)tanh(v/v_max);
}

inline vector_3 ContinuousMaxMagnitude( vector_3 const &v, float32 v_max )
{
   float32 v_l = v.Length();
   float32 v_m = 1.f;
   if(v_l > 1e-6)
   {
      v_m = ContinuousMaxMagnitude(v_l,v_max) / v_l;
   }
   return v_m*v;
}

inline float32 vmaxel( vector_3 const &a )
{
   if(a(0) > a(1)) {
      if(a(2) > a(0)) {
         return a(2);
      } else {
         return a(0);
      }
   } else {
      if(a(2) > a(1)) {
         return a(2);
      } else {
         return a(1);
      }
   }
}
inline float32 vminel( vector_3 const &a )
{
   if(a(0) < a(1)) {
      if(a(2) < a(0)) {
         return a(2);
      } else {
         return a(0);
      }
   } else {
      if(a(2) < a(1)) {
         return a(2);
      } else {
         return a(1);
      }
   }
}
   
inline vector_3 vmax( vector_3 const &a, vector_3 const &b )
{
   vector_3 m(vector_3::DoNotInitialize);
   for(int i = 0;i < 3;++i) {
      m(i) = a(i) > b(i) ? a(i) : b(i);
   }
   return m;
}
inline vector_3 vmin( vector_3 const &a, vector_3 const &b )
{
   vector_3 m(vector_3::DoNotInitialize);
   for(int i = 0;i < 3;++i) {
      m(i) = a(i) < b(i) ? a(i) : b(i);
   }
   return m;
}
inline bool v_lt( vector_3 const &a, vector_3 const &b )
{
   return (a(0) < b(0)) && (a(1) < b(1)) && (a(2) < b(2));
}
inline bool v_lte( vector_3 const &a, vector_3 const &b )
{
   return (a(0) <= b(0)) && (a(1) <= b(1)) && (a(2) <= b(2));
}
inline bool v_gt( vector_3 const &a, vector_3 const &b )
{
   return (a(0) > b(0)) && (a(1) > b(1)) && (a(2) > b(2));
}
inline bool v_gte( vector_3 const &a, vector_3 const &b )
{
   return (a(0) >= b(0)) && (a(1) >= b(1)) && (a(2) >= b(2));
}

inline matrix_3x3 RotationAtoBNormalized( vector_3 const &a, vector_3 const &b )
{
   float32 c = DotProduct(a,b);
   vector_3 abn = CrossProduct(a,b);
   float32 l = 0.0;
   vector_3 n = Normalize(abn,&l);
   if(l > 0.0) {
      float32 s = l;
      return matrix_3x3(n,c,s);
   } else {
      if(c > 0.0) {  // no rotation needed
         return matrix_3x3();
      } else {       // 180 rotation needed
         vector_3 y,z;
         GetDextralCoordinateSystem(b-a,y,z);
         return matrix_3x3(y,-1,0);
      }
   }
}
inline matrix_3x3 RotationAtoB( vector_3 const &a, vector_3 const &b )
{
   return RotationAtoBNormalized(Normalize(a),Normalize(b));
}

inline bool IsInBounds_IncInc( vector_3 const &p, vector_3 const &mn, vector_3 const &mx )
{
   return v_gte(p,mn) && v_lte(p,mx);
}
inline bool IsInBounds_IncExc( vector_3 const &p, vector_3 const &mn, vector_3 const &mx )
{
   return v_gte(p,mn) && v_lt(p,mx);
}
inline bool IsInBounds_ExcInc( vector_3 const &p, vector_3 const &mn, vector_3 const &mx )
{
   return v_gt(p,mn) && v_lte(p,mx);
}
inline bool IsInBounds_ExcExc( vector_3 const &p, vector_3 const &mn, vector_3 const &mx )
{
   return v_gt(p,mn) && v_lt(p,mx);
}

#pragma warning (disable:4514)		// unreferenced inline removed

#define MATH3D_H
#endif

